/*
** Adpcm - Sources\Commun\Adpcm\Encoder\Adpcm.c
** Nicolas ROBERT [NRX] - Hong Kong 2005

** GUI Version - Arnaud BOUR [NAGAME] - France 2005
*/

//////////////
// WaveInit //
//////////////
unsigned char WaveInit(Wave* wave)
{
   unsigned char buffer[12];
   unsigned char formatOk;
   unsigned long length;

   // Read the header
   if(fread(buffer,sizeof(unsigned char),12,wave->file)<12)
      return(1);

   // Check for the RIFF & WAVE signatures
   if(memcmp("RIFF",buffer,4)||memcmp("WAVE",buffer+8,4))
      return(1);

   // Parse the chunks
   formatOk=0;
   while(fread(buffer,sizeof(unsigned char),4,wave->file))
   {
      fread(&length,sizeof(unsigned long),1,wave->file);
      if(!memcmp("fmt ",buffer,4))
      {
         // Read the format
         if(length<16)
            return(1);

         fread(&wave->format,sizeof(wave->format),1,wave->file);
         fseek(wave->file,length-sizeof(wave->format),SEEK_CUR);
         formatOk=1;

         // Compute the "bytesPerSample" from the "bitsPerSample"
         wave->bytesPerSample=(wave->format.bitsPerSample+7)>>3;
      }
      else if(!memcmp("data",buffer,4))
      {
         // We now have some data; first check we got the format
         if(!formatOk)
            return(1);

         // Compute the actual number of samples
         wave->numberSamples=length/(wave->format.numberChannels*wave->bytesPerSample);

         // Everything seems ok!
         return(0);
      }
      else
      {
         // Skip unrecognized chunk types
         fseek(wave->file,length,SEEK_CUR);
      }
   }

   // We haven't found the data!
   return(1);
}

/////////////////
// AdpcmEncode //
/////////////////
void AdpcmEncode(Adpcm* adpcm,signed short* source,unsigned char* dest,unsigned short length)
{
   signed long sample;
   signed short index;
   signed short step;
   unsigned char byte;
   unsigned char code;
   unsigned char sign;
   signed short diff;

   // Retrieve the previous data
   sample=adpcm->lastSample;
   index=adpcm->lastIndex;

   // Process the samples
   byte=0;
   while(length)
   {
      // Get the step
      step=adpcmStep[index];

      // Compute the code
      diff=(*source++)-sample;
      if(diff<0)
      {
         diff=-diff;
         sign=8;
      }
      else
         sign=0;
      code=(diff<<2)/step;
      if(code>7)
         code=7;
	  else if(OptimisedMode && code==7)
         code=6;

      // Modify the previous sample
      diff=step>>3;
      if(code&1)
         diff+=step>>2;
      if(code&2)
         diff+=step>>1;
      if(code&4)
         diff+=step;
      if(sign)
      {
         sample-=diff;
         if(sample<-32768)
            sample=-32768;
      }
      else
      {
         sample+=diff;
         if(sample>32767)
            sample=32767;
      }

      // Compute the next index
      index+=adpcmModifyIndex[code];
      if(index<0)
         index=0;
      else if(index>88)
         index=88;

     // Store the sample
      code|=sign;
      if(length&1)
         *dest++=(code<<4)|byte;
      else
         byte=code;

      // Next sample...
      --length;
   }

   // Store the current data
   adpcm->lastSample=sample;
   adpcm->lastIndex=index;
}

////////////////
// convertion //
////////////////
int Convertir(HWND hwnd)
{
   unsigned long i;

   char message[256];
   unsigned char conv;
   unsigned long originalNumberSamples;
   unsigned long adpcmFileSize;


   Wave wave;
   unsigned char waveShiftValue;
   unsigned char waveCounter;
   unsigned short waveChannel;
   signed long waveSumSamples;
   signed long waveLongSample;
   signed short waveSamples[WAVE_SAMPLES];

   Adpcm adpcm;
   unsigned long adpcmLength;
   unsigned char adpcmCounter;
   unsigned char adpcmSamples[ADPCM_SAMPLES];

   // Init Progress Bar
   SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, (WPARAM) 0, 0);

   // Check the arguments
   noCompress=IsDlgButtonChecked(hwnd, IDC_NOCOMPRESS);
   OptimisedMode=IsDlgButtonChecked(hwnd, IDC_OPTIMISED);


   // Open & check the wave file
   if(!strlen(FileTitle))
   {
      MessageBox(NULL, "Select the wave file", "Stop !" ,MB_OK | MB_ICONWARNING);
	  return(1);
   }
	  
   // Open DialogBox
   if (!Sauvegarde(hwnd)) return(1);

   // Binaire format ?
   if (!strcmp(FileFormatSave, "bin")) binFormat=TRUE;


   wave.file=fopen(FileName, "rb");

   if(!wave.file)
   {
      MessageBox(NULL, "Couldn't open the wave file", "Error" ,MB_OK | MB_ICONERROR);
      return(1);
   }
   if(WaveInit(&wave))
   {
      MessageBox(NULL, "Incorrect wave file", "Error" ,MB_OK | MB_ICONERROR);
      fclose(wave.file);
      return(1);
   }
   if(!noCompress && wave.format.sampleRate>MAX_WAVE_FREQUENCY)
   {
	  sprintf(message, "The sample rate is greater than %dHz", MAX_WAVE_FREQUENCY);
      MessageBox(NULL, message, "Error" ,MB_OK | MB_ICONERROR);      

      fclose(wave.file);
      return(1);
   }

   // Open the adpcm file
   if(binFormat) adpcm.file=fopen(FileNameSave,"wb");
   else adpcm.file=fopen(FileNameSave,"w");

   if(!adpcm.file)
   {
      MessageBox(NULL, "Couldn't write to adpcm file", "Error" ,MB_OK | MB_ICONERROR); 

      fclose(wave.file);
      return(1);
   }

   // Copy NumberSamples for ProgressBar
   originalNumberSamples=wave.numberSamples;

   // Compute the number of data in our adpcm file
   if(noCompress)
      adpcmLength=wave.numberSamples;
   else
   {
      adpcmLength=(wave.numberSamples+1)>>1;

	  // Pour la verif d'alignement
	  adpcmFileSize=adpcmLength;

      // Also, initialize the adpcm
      adpcm.lastSample=0;
      adpcm.lastIndex=0;
   }

   // Compute the number of bits the value will need to be shifted to fit on 16 bits
   waveShiftValue=(wave.bytesPerSample-2)<<3;

   // Write the header of our adpcm file
   if(binFormat)
   {
	  //fwrite(&FileTitle,   strlen(FileTitle), 1, adpcm.file);
	  fwrite(&wave.format.sampleRate,  sizeof(unsigned long), 1, adpcm.file);
	  fwrite(&adpcmLength, sizeof(unsigned long), 1, adpcm.file);

	  conv=noCompress;
	  fwrite(&conv,  sizeof(unsigned long), 1, adpcm.file);  

   } else 
	  fprintf(adpcm.file,
        ".section .rodata\n.global ADPCM_%s\n.align 2\nADPCM_%s:\n.word %lu\n.word %lu\n.byte %d\n.space 3",
        FileTitle,
        FileTitle,
        wave.format.sampleRate,
        adpcmLength,
        noCompress);

   // Update Progress Bar
   SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, (WPARAM) 25, 0);

   // Process the wave file...
   while(wave.numberSamples)
   {
      // Read some samples
      for(waveCounter=0;waveCounter<WAVE_SAMPLES && wave.numberSamples;++waveCounter,--wave.numberSamples)
      {
         // We compute the average of all the channels and put the result on 16 bits
         waveSumSamples=0;
         for(waveChannel=wave.format.numberChannels;waveChannel;--waveChannel)
         {
            if(wave.bytesPerSample==1)
               waveSumSamples+=(signed short)((fgetc(wave.file)<<8)-32768);
            else
            {
               fread(&waveLongSample,wave.bytesPerSample,1,wave.file);
               waveSumSamples+=(signed short)(waveLongSample>>waveShiftValue);
            }
         }
         waveSamples[waveCounter]=waveSumSamples/wave.format.numberChannels;
      }

      // Let's put these data in the output file
      if(noCompress)
      {
         // Directly save the samples to the file (no compression)
         if(binFormat)
		 {
            for(adpcmCounter=0;adpcmCounter<waveCounter;++adpcmCounter)
			{
			   conv=waveSamples[adpcmCounter]>>8;
               fwrite(&conv, sizeof(unsigned char), 1, adpcm.file);
			}

		 } else {

            fprintf(adpcm.file,"\n.byte %d",(unsigned char)(waveSamples[0]>>8));
            for(adpcmCounter=1;adpcmCounter<waveCounter;++adpcmCounter)
               fprintf(adpcm.file,",%d",(unsigned char)(waveSamples[adpcmCounter]>>8));
		 }
      }
      else
      {
         // Be sure we have an event number of data
         if(waveCounter&1)
            waveSamples[waveCounter++]=0;

         // Encoding...
         AdpcmEncode(&adpcm,waveSamples,adpcmSamples,waveCounter);

         // Save the compressed data
         waveCounter>>=1;

         if(binFormat)
		 {
            for(adpcmCounter=0;adpcmCounter<waveCounter;++adpcmCounter)
			    fwrite(&adpcmSamples[adpcmCounter], sizeof(unsigned char), 1, adpcm.file);

		 } else {
            fprintf(adpcm.file,"\n.byte %d",adpcmSamples[0]);
            for(adpcmCounter=1;adpcmCounter<waveCounter;++adpcmCounter)
                fprintf(adpcm.file,",%d",adpcmSamples[adpcmCounter]);
		 }
      }

      // Update the "adpcmLength" to later check everything was ok...
      adpcmLength-=waveCounter;

	  // Update Progress Bar
	  SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, (WPARAM) (((originalNumberSamples-wave.numberSamples)/originalNumberSamples)*75), 0);
   }

   // V1.1 : On s'assure que le fichier resultant est sur un alignement 32bit
   conv=0;

   if(binFormat)
   {
      adpcmFileSize=(floor( ftell(adpcm.file) /4 ) +1) *4;
      
      while(ftell(adpcm.file)<adpcmFileSize)
         fwrite(&conv, sizeof(unsigned char), 1, adpcm.file);

   } else {

      adpcmFileSize=(floor( (adpcmFileSize+12)/4 ) +1) *4;

	  for(i=adpcmFileSize; i<=adpcmFileSize; i++)
	  	 fprintf(adpcm.file,",%d", conv);
   }

   // Close the files
   if(!binFormat) fputc('\n',adpcm.file);


   fclose(adpcm.file);
   fclose(wave.file);

   // Update Progress Bar
   SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, (WPARAM) 100, 0);

   // Just a final little check!
   if(adpcmLength)
   {
	  sprintf(message, "%lu bytes missing", adpcmLength);
      MessageBox(NULL, message, "Error" ,MB_OK | MB_ICONERROR);
	  
   } else {
	  sprintf(message, "%s convert with success !", FileTitle);
      MessageBox(NULL, message, "Successful :)" ,MB_OK | MB_ICONINFORMATION);
   }

   // Bye bye...
   return(0);
}
